use crate::parser::Instruction;

pub struct Interpreter {
  pointer: isize,
  cells: std::collections::HashMap<isize, u8>,
}

impl Interpreter {
  pub fn new() -> Interpreter {
    Interpreter {
      pointer: 0,
      cells: std::collections::HashMap::new(),
    }
  }

  pub fn run(&mut self, instructions: &Vec<Instruction>) {
    for instruction in instructions {
      self.run_instruction(instruction);
    }
  }

  fn run_instruction(&mut self, instruction: &Instruction) {
    match instruction {
      Instruction::IncrementPointer => self.increment_pointer(),
      Instruction::DecrementPointer => self.decrement_pointer(),
      Instruction::IncrementValue => self.increment_value(),
      Instruction::DecrementValue => self.decrement_value(),
      Instruction::OutputValue => self.output_value(),
      Instruction::ReadValue => self.read_value(),
      Instruction::Branch(body) => self.branch(&body),
    }
  }

  fn increment_pointer(&mut self) {
    self.pointer += 1;
  }

  fn decrement_pointer(&mut self) {
    self.pointer -= 1;
  }

  fn increment_value(&mut self) {
    self.set_current_cell(self.current_cell().wrapping_add(1))
  }

  fn decrement_value(&mut self) {
    self.set_current_cell(self.current_cell().wrapping_sub(1))
  }

  fn output_value(&mut self) {
    print!("{}", self.current_cell() as char);
  }

  fn read_value(&mut self) {
  }

  fn branch(&mut self, body: &Vec<Instruction>) {
    while self.current_cell() != 0 {
      self.run(body);
    }
  }

  fn current_cell(&self) -> u8 {
    *self.cells.get(&self.pointer).unwrap_or(&0)
  }

  fn set_current_cell(&mut self, value: u8) {
    self.cells.insert(self.pointer, value);
  }
}
